/**
 * JORM: an implementation of a generic mapping system for persistent Java
 * objects. Two mapping are supported: to RDBMS and to binary files. Copyright
 * (C) 2001-2004 France Telecom R&D - INRIA
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) any
 * later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Contact: jorm@objectweb.org
 *  
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectStreamClass;
import java.io.Reader;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Array;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Date;
import java.util.List;

import org.objectweb.jorm.mapper.rdb.adapter.api.JoinedTable;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import org.objectweb.jorm.mapper.rdb.adapter.api.TypeAdapter;
import org.objectweb.jorm.mapper.rdb.lib.RdbBlob;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/*
 * This class is a JDBC compliant implementation of the RdbAdapter interface.
 * That means a compliant JDBC driver should work with this implementation. When
 * a JDBC driver is not conform for some point, it is adviced to sub class this
 * implementation for redefining the functions which do no work.
 * 
 * @author Sebastien Chassande-Barrioz
 */
public class BasicRdbAdapter implements RdbAdapter {

	private final static String CONCAT = "concat";

	private final static String INDEXEDLOCATE = "instr";
	
	protected Logger logger;

	/**
	 * Is the name of the RdbAdapter. The name is used, in particular, in the
	 * RdbAdapter factory for choosing an adapter.
	 */
	protected String name;

	public BasicRdbAdapter() {
	}

	public BasicRdbAdapter(String _name) {
		this.name = _name;
	}

	public boolean supportBatchPreparedStatement() {
		return true;
	}

	public String getName() {
		return name;
	}

	/**
	 * This methods return the SQL type code linked to the java type specified
	 * in parameter.
	 * 
	 * @return An int constant defined in the java.sql.Types class
	 */
	public int getSqlTypeCode(int typeCode) throws RdbAdapterException {
		return getSqlTypeCode(typeCode, null);
	}

	/**
	 * This methods return the SQL type code linked to the java type specified
	 * in parameter.
	 * 
	 * @return An int constant defined in the java.sql.Types class
	 */
	public int getSqlTypeCode(int typeCode, String sqlColumnType)
			throws RdbAdapterException {
        if (BasicArray.isArray(typeCode)) {
            return Types.ARRAY;
        }
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return Types.BIT;
		case TYPECODE_CHAR:
		case TYPECODE_OBJCHAR:
			return Types.CHAR;
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
			return Types.TINYINT;
		case TYPECODE_SHORT:
		case TYPECODE_OBJSHORT:
			return Types.SMALLINT;
		case TYPECODE_INT:
		case TYPECODE_OBJINT:
			return Types.INTEGER;
		case TYPECODE_LONG:
		case TYPECODE_OBJLONG:
			return Types.BIGINT;
		case TYPECODE_FLOAT:
		case TYPECODE_OBJFLOAT:
			return Types.REAL;
		case TYPECODE_DOUBLE:
		case TYPECODE_OBJDOUBLE:
			return Types.DOUBLE;
		case TYPECODE_STRING:
			return Types.VARCHAR;
		case TYPECODE_DATE:
			if (sqlColumnType == null) {
				return Types.TIMESTAMP;
			} else {
				String str = sqlColumnType.toUpperCase();
				if (str.startsWith("DATE")) {
					return Types.DATE;
				} else if (str.startsWith("TIMESTAMP")) {
					return Types.TIMESTAMP;
				} else if (str.startsWith("TIME")) {
					return Types.TIME;
				} else {
					return Types.TIMESTAMP;
				}
			}
		case TYPECODE_CHARARRAY:
			return Types.VARCHAR;
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
			return Types.VARBINARY;
		case TYPECODE_BIGINTEGER:
			return Types.DECIMAL;
		case TYPECODE_BIGDECIMAL:
			return Types.DECIMAL;
		default:
			throw new RdbAdapterException(
					"No sql type exist for the type code " + typeCode);
		}
	}

	public String getSqlType(int typeCode, boolean usedInPK, int size, int scale)
			throws RdbAdapterException {
        if (BasicArray.isArray(typeCode)) {
            return getSqlType(typeCode - TYPECODE_ARRAY, usedInPK, size, scale) + "[]";
        }
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return "BIT";
		case TYPECODE_CHAR:
		case TYPECODE_OBJCHAR:
			return "CHAR(1)";
		case TYPECODE_BYTE:
		case TYPECODE_OBJBYTE:
			return "TINYINT";
		case TYPECODE_SHORT:
		case TYPECODE_OBJSHORT:
			return "SMALLINT";
		case TYPECODE_INT:
		case TYPECODE_OBJINT:
			return "INTEGER";
		case TYPECODE_LONG:
		case TYPECODE_OBJLONG:
			return "BIGINT";
		case TYPECODE_FLOAT:
		case TYPECODE_OBJFLOAT:
			return "REAL";
		case TYPECODE_DOUBLE:
		case TYPECODE_OBJDOUBLE:
			return "DOUBLE";
		case TYPECODE_STRING:
		case TYPECODE_CHARARRAY:
			if (size != NOSIZE) {
				return "VARCHAR(" + size + ")";
			}
			return "VARCHAR";
		case TYPECODE_DATE:
			return "TIMESTAMP";
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
			if (size != NOSIZE) {
				return "VARBINARY(" + size + ")";
			}
			return "VARBINARY";
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			if (size != NOSIZE) {
				if (scale != NOSIZE) {
					return "NUMERIC(" + size + ", " + scale + ")";
				} else {
					return "NUMERIC(" + size + ")";
				}
			} else {
				return "NUMERIC";
			}
		default:
			throw new RdbAdapterException("No sql type exist for the type code"
					+ typeCode);
		}
	}

	public void setNull(PreparedStatement ps, int idx, int sqlType)
			throws SQLException {
		ps.setNull(idx, sqlType);
	}

	/**
	 * This calculates the size of a result set. If this feature is not
	 * supported by a database or its jdbc driver, this method returns -1.
	 * 
	 * @param rs the result set which the size must be returned
	 */
	public int fetchResultSetSize(ResultSet rs) throws SQLException {
		return -1;
	}

	public String getValueAsSQLString(Object value, int typeCode) {
        if (BasicArray.isArray(typeCode)) {
        	String res = getArrayValueAsSQLString(value, typeCode);
            return res;
        }
		switch (typeCode) {
		case TYPECODE_BOOLEAN:
		case TYPECODE_OBJBOOLEAN:
			return (((Boolean) value).booleanValue() ? "1" : "0");

		case TYPECODE_OBJBYTE:
		case TYPECODE_OBJSHORT:
		case TYPECODE_OBJINT:
		case TYPECODE_OBJLONG:
		case TYPECODE_OBJFLOAT:
		case TYPECODE_OBJDOUBLE:
		case TYPECODE_BYTE:
		case TYPECODE_SHORT:
		case TYPECODE_INT:
		case TYPECODE_LONG:
		case TYPECODE_FLOAT:
		case TYPECODE_DOUBLE:
			return value.toString();

		case TYPECODE_OBJCHAR:
		case TYPECODE_STRING:
		case TYPECODE_DATE:
		case TYPECODE_BIGINTEGER:
		case TYPECODE_BIGDECIMAL:
			return "'" + value.toString() + "'";

		case TYPECODE_CHARARRAY:
		case TYPECODE_BYTEARRAY:
		case TYPECODE_SERIALIZED:
		default:
			return null;
		}
	}
	
	public String getArrayToStringBegin() {
		return "{";
	}
	
	public String getArrayToStringEnd() {
		return "}";
	}
	
	public String getArrayToStringSeparator() {
		return ",";
	}
	
	public String getArrayToString(Object array, int typeCode) {
		return getArrayValueAsSQLString(array, typeCode, 
				getArrayToStringBegin(),
				getArrayToStringSeparator(),
				getArrayToStringEnd());
	}
	
	public String getArrayValueAsSQLStringBegin() {
		return "'{";
	}
	
	public String getArrayValueAsSQLStringEnd() {
		return "}'";
	}
	
	public String getArrayValueAsSQLStringSeparator() {
		return ",";
	}

	public String getArrayElementAsSQLString(Object elem, int elemTypeCode) {
		return getValueAsSQLString(elem, elemTypeCode);
	}

	public String getArrayValueAsSQLString(Object array, int typeCode) {
		return getArrayValueAsSQLString(array, typeCode, 
				getArrayValueAsSQLStringBegin(),
				getArrayValueAsSQLStringSeparator(),
				getArrayValueAsSQLStringEnd());
	}	
	public String getArrayValueAsSQLString(Object array, 
			int typeCode,
			String begin,
			String realsep,
			String end) {
        if (array == null) {
            return "NULL";
        } else {
        	int innerTypeCode = BasicArray.getElementArrayTypeCode(typeCode);
	        StringBuffer sb = new StringBuffer();
	        sb.append(begin);
	        String sep = "";
            switch(innerTypeCode) {
            case TypeAdapter.TYPECODE_OBJBOOLEAN:
            case TypeAdapter.TYPECODE_OBJBYTE:
            case TypeAdapter.TYPECODE_OBJSHORT:
            case TypeAdapter.TYPECODE_OBJINT:
            case TypeAdapter.TYPECODE_OBJLONG:
            case TypeAdapter.TYPECODE_OBJFLOAT:
            case TypeAdapter.TYPECODE_OBJDOUBLE:
            case TypeAdapter.TYPECODE_DATE:
            case TypeAdapter.TYPECODE_BIGDECIMAL:
            case TypeAdapter.TYPECODE_BIGINTEGER:
            case TypeAdapter.TYPECODE_OBJCHAR:
            case TypeAdapter.TYPECODE_STRING:
            {
            	Object[] tab = (Object[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getArrayElementAsSQLString(tab[i], innerTypeCode));
                }
                break;
            }
            case TypeAdapter.TYPECODE_BOOLEAN: 
            {
                boolean[] tab = (boolean[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_BYTE: 
            {
                byte[] tab = (byte[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_CHAR: 
            {
                char[] tab = (char[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_SHORT: 
            {
                short[] tab = (short[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_INT: 
            {
                int[] tab = (int[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_LONG: 
            {
                long[] tab = (long[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_FLOAT: 
            {
                float[] tab = (float[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            case TypeAdapter.TYPECODE_DOUBLE: 
            {
                double[] tab = (double[]) array;
                for (int i = 0; i < tab.length; i++) {
                	sb.append(sep);
                	sep = realsep;
                    sb.append(getValueAsSQLString(tab[i]));
                }
                break;
            }
            
            }
	        sb.append(end);
	        return sb.toString();
        }
		
	}

	public String getValueAsSQLString(boolean value) {
		StringBuffer sb = new StringBuffer();
		if (value) {
			return sb.append("1").toString();
		} else {
			return sb.append("0").toString();
		}
	}

	public String getValueAsSQLString(byte value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public String getValueAsSQLString(char value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("'").append(value).append("'").toString();
	}

	public String getValueAsSQLString(short value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public String getValueAsSQLString(int value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public String getValueAsSQLString(long value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public String getValueAsSQLString(float value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public String getValueAsSQLString(double value) {
		StringBuffer sb = new StringBuffer();
		return sb.append("").append(value).toString();
	}

	public boolean getBoolean(ResultSet rs, int idx, boolean nullValue)
			throws SQLException {
		boolean res = rs.getBoolean(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Boolean getOboolean(ResultSet rs, int idx, Boolean nullValue)
			throws SQLException {
		boolean res = rs.getBoolean(idx);
		return (rs.wasNull() ? null : new Boolean(res));
	}

	public char getChar(ResultSet rs, int idx, char nullValue)
			throws SQLException {
		char res = rs.getString(idx).charAt(0);
		return (rs.wasNull() ? nullValue : res);
	}

	public Character getOchar(ResultSet rs, int idx, Character nullValue)
			throws SQLException {
		String s = rs.getString(idx);
		return (rs.wasNull() ? nullValue : new Character(s.charAt(0)));
	}

	public byte getByte(ResultSet rs, int idx, byte nullValue)
			throws SQLException {
		byte res = rs.getByte(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Byte getObyte(ResultSet rs, int idx, Byte nullValue)
			throws SQLException {
		byte res = rs.getByte(idx);
		return (rs.wasNull() ? null : new Byte(res));
	}

	public short getShort(ResultSet rs, int idx, short nullValue)
			throws SQLException {
		short res = rs.getShort(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Short getOshort(ResultSet rs, int idx, Short nullValue)
			throws SQLException {
		short res = rs.getShort(idx);
		return (rs.wasNull() ? nullValue : new Short(res));
	}

	public int getInt(ResultSet rs, int idx, int nullValue) throws SQLException {
		int res = rs.getInt(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Integer getOint(ResultSet rs, int idx, Integer nullValue)
			throws SQLException {
		int res = rs.getInt(idx);
		return (rs.wasNull() ? nullValue : new Integer(res));
	}

	public long getLong(ResultSet rs, int idx, long nullValue)
			throws SQLException {
		long res = rs.getLong(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Long getOlong(ResultSet rs, int idx, Long nullValue)
			throws SQLException {
		long res = rs.getLong(idx);
		return (rs.wasNull() ? nullValue : new Long(res));
	}

	public float getFloat(ResultSet rs, int idx, float nullValue)
			throws SQLException {
		float res = rs.getFloat(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Float getOfloat(ResultSet rs, int idx, Float nullValue)
			throws SQLException {
		float res = rs.getFloat(idx);
		return (rs.wasNull() ? null : new Float(res));
	}

	public double getDouble(ResultSet rs, int idx, double nullValue)
			throws SQLException {
		double res = rs.getDouble(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Double getOdouble(ResultSet rs, int idx, Double nullValue)
			throws SQLException {
		double res = rs.getDouble(idx);
		return (rs.wasNull() ? nullValue : new Double(res));
	}

	public String getString(ResultSet rs, int idx, String nullValue)
			throws SQLException {
		String res = rs.getString(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Date getDate(ResultSet rs, int idx, Date nullValue)
			throws SQLException {
		return getDate(rs, idx, null, nullValue);
	}

	/*
	 * before Java 1.4 Timstamp#getTime() returned only the seconds value part
	 * of the timestamp and not the milliseconds which are record in the nanos
	 * part of the timestamp
	 */
	private static final boolean accountForNanos;

	static {
		String version = System.getProperty("java.specification.version");
		if (version.startsWith("1.")) {
			int minor = new Integer(version.charAt(2)).intValue();
			if (minor < 4) {
				accountForNanos = true;
			} else {
				accountForNanos = false;
			}
		} else {
			accountForNanos = false;
		}
	}

	private static Date timestamp2date(java.sql.Timestamp tstamp) {
		if (tstamp == null) {
			return null;
		}
		if (accountForNanos) {
			return new java.util.Date(tstamp.getTime() + tstamp.getNanos()
					/ 1000000);
		} else {
			return new java.util.Date(tstamp.getTime());
		}
	}

	private static Date sqldate2date(java.sql.Date d) {
		if (d == null) {
			return null;
		}
		return new java.util.Date(d.getTime());
	}

	private static Date time2date(java.sql.Time t) {
		if (t == null) {
			return null;
		}
		return new java.util.Date(t.getTime());
	}

	public Date getDate(ResultSet rs,
			int idx,
			String sqlColumnType,
			Date nullValue) throws SQLException {
		Date res = null;
		if (sqlColumnType == null || sqlColumnType.length() == 0) {
			res = timestamp2date(rs.getTimestamp(idx));
		} else {
			String str = sqlColumnType.toUpperCase();
			if (str.startsWith("DATE")) {
				res = sqldate2date(rs.getDate(idx));
			} else if (str.startsWith("TIMESTAMP")) {
				res = timestamp2date(rs.getTimestamp(idx));
			} else if (str.startsWith("TIME")) {
				res = time2date(rs.getTime(idx));
			} else {
				res = timestamp2date(rs.getTimestamp(idx));
			}
		}
		if (rs.wasNull()) {
			return nullValue;
		} else {
			return res;
		}
	}

	public char[] getCharArray(ResultSet rs, int idx, char[] nullValue)
			throws SQLException {
		Reader r = rs.getCharacterStream(idx);
		if (rs.wasNull()) {
			return nullValue;
		} else {
			try {
				return getCharArray(r);
			} catch (IOException e) {
				throw new SQLException("Impossible to build a char[]: "
						+ e.getMessage());
			}
		}
	}

	private static char[] getCharArray(java.io.Reader reader)
			throws java.io.IOException {
		int size = 0;
		long skipped = 0;
		while ((skipped = reader.skip(50)) == 50) {
			size += 50;
		}
		size += (int) skipped;
		char[] res = new char[size];
		reader.reset();
		reader.read(res);
		return res;
	}

	public byte[] getByteArray(ResultSet rs, int idx, byte[] nullValue)
			throws SQLException {
		byte[] res = rs.getBytes(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public Serializable getSerialized(ResultSet rs,
			int idx,
			Serializable nullValue) throws SQLException, IOException,
			ClassNotFoundException {
	    class Specialized_OIS extends ObjectInputStream { 
	        Specialized_OIS(InputStream is) throws IOException { 
	            super(is); 
	        } 
	        
	        protected Class resolveClass(ObjectStreamClass osc) throws IOException, ClassNotFoundException { 
	            try { 
	                return super.resolveClass(osc); 
	            } catch (ClassNotFoundException e) { 
	                return Thread.currentThread().getContextClassLoader()
	                .loadClass(osc.getName()); 
	            } 
	        } 
	    } 
        InputStream is = rs.getBinaryStream(idx); 
        return (rs.wasNull() 
                ? nullValue
                : (Serializable) new Specialized_OIS(is).readObject()); 
	}

	public BigDecimal getBigDecimal(ResultSet rs, int idx, BigDecimal nullValue)
			throws SQLException {
		BigDecimal res = rs.getBigDecimal(idx);
		return (rs.wasNull() ? nullValue : res);
	}

	public BigInteger getBigInteger(ResultSet rs, int idx, BigInteger nullValue)
			throws SQLException {
		BigDecimal bd = getBigDecimal(rs, idx, null);
		if (bd == null) {
			return nullValue;
		} else {
			return bd.toBigInteger();
		}
	}

	public Object getArray(ResultSet rs, int idx, Object nullValue, int typeCode)
			throws SQLException {
		Array a = rs.getArray(idx);
		if (a == null || rs.wasNull()) {
			return nullValue;
		}
		Object v = a.getArray();
		if (v == null) {
			return nullValue;
		}
		switch(BasicArray.getElementArrayTypeCode(typeCode)) {
        case TypeAdapter.TYPECODE_OBJBOOLEAN:
        {
        	boolean[] tab = (boolean[]) v;
        	Boolean[] res = new Boolean[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = Boolean.valueOf(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJBYTE:
        {
        	byte[] tab = (byte[]) v;
        	Byte[] res = new Byte[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Byte(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJSHORT:
        {
        	short[] tab = (short[]) v;
        	Short[] res = new Short[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Short(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJINT:
        {
        	int[] tab = (int[]) v;
        	Integer[] res = new Integer[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Integer(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJLONG:
        {
        	long[] tab = (long[]) v;
        	Long[] res = new Long[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Long(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJFLOAT:
        {
        	short[] tab = (short[]) v;
        	Short[] res = new Short[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Short(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJDOUBLE:
        {
        	double[] tab = (double[]) v;
        	Double[] res = new Double[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Double(tab[i]);
        	}
        	return res;
        }	
        case TypeAdapter.TYPECODE_OBJCHAR:
        {
        	char[] tab = (char[]) v;
        	Character[] res = new Character[tab.length];
        	for(int i=0; i<tab.length; i++) {
        		res[i] = new Character(tab[i]);
        	}
        	return res;
        }	
		default:
			return v;
		}
	}

	public void setBoolean(PreparedStatement ps, int idx, boolean o)
			throws SQLException {
		ps.setBoolean(idx, o);
	}

	public void setOboolean(PreparedStatement ps, int idx, Boolean o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.BIT);
		else
			ps.setBoolean(idx, o.booleanValue());
	}

	public void setChar(PreparedStatement ps, int idx, char c)
			throws SQLException {
		ps.setString(idx, String.valueOf(c));
	}

	public void setOchar(PreparedStatement ps, int idx, Character o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.CHAR);
		else {
			ps.setString(idx, String.valueOf(o.charValue()));
		}
	}

	public void setByte(PreparedStatement ps, int idx, byte o)
			throws SQLException {
		ps.setByte(idx, o);
	}

	public void setObyte(PreparedStatement ps, int idx, Byte o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.SMALLINT);
		else
			ps.setByte(idx, o.byteValue());
	}

	public void setShort(PreparedStatement ps, int idx, short o)
			throws SQLException {
		ps.setShort(idx, o);
	}

	public void setOshort(PreparedStatement ps, int idx, Short o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.SMALLINT);
		else
			ps.setShort(idx, o.shortValue());
	}

	public void setInt(PreparedStatement ps, int idx, int o)
			throws SQLException {
		ps.setInt(idx, o);
	}

	public void setOint(PreparedStatement ps, int idx, Integer o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.INTEGER);
		else
			ps.setInt(idx, o.intValue());
	}

	public void setLong(PreparedStatement ps, int idx, long o)
			throws SQLException {
		ps.setLong(idx, o);
	}

	public void setOlong(PreparedStatement ps, int idx, Long o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.BIGINT);
		else
			ps.setLong(idx, o.longValue());
	}

	public void setFloat(PreparedStatement ps, int idx, float o)
			throws SQLException {
		ps.setFloat(idx, o);
	}

	public void setOfloat(PreparedStatement ps, int idx, Float o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.FLOAT);
		else
			ps.setFloat(idx, o.floatValue());
	}

	public void setDouble(PreparedStatement ps, int idx, double o)
			throws SQLException {
		ps.setDouble(idx, o);
	}

	public void setOdouble(PreparedStatement ps, int idx, Double o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DOUBLE);
		else
			ps.setDouble(idx, o.doubleValue());
	}

	public void setString(PreparedStatement ps, int idx, String o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARCHAR);
		else
			ps.setString(idx, o);
	}

	public void setDate(PreparedStatement ps, int idx, Date o, String columnType)
			throws SQLException {
		if (columnType == null) {
			//default is a Timestamp
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIMESTAMP);
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("DATE")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.DATE);
			else if (o instanceof java.sql.Date) {
				ps.setDate(idx, (java.sql.Date) o);
			} else {
				ps.setDate(idx, new java.sql.Date(o.getTime()));
			}
		} else if (columnType.startsWith("TIMESTAMP")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIMESTAMP);
			else if (o instanceof java.sql.Timestamp) {
				ps.setTimestamp(idx, (java.sql.Timestamp) o);
			} else {
				ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
			}
		} else if (columnType.startsWith("TIME")) {
			if (o == null)
				ps.setNull(idx, java.sql.Types.TIME);
			else if (o instanceof java.sql.Time) {
				ps.setTime(idx, (java.sql.Time) o);
			} else {
				ps.setTime(idx, new java.sql.Time(o.getTime()));
			}
		}
	}

	public void setCharArray(PreparedStatement ps, int idx, char[] o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else {
			ps
					.setCharacterStream(idx, new java.io.CharArrayReader(o),
							o.length);
		}
	}

	public void setByteArray(PreparedStatement ps, int idx, byte[] o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else
			ps.setBytes(idx, o);
	}

	public void setBigInteger(PreparedStatement ps, int idx, BigInteger o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DECIMAL);
		else
			ps.setBigDecimal(idx, new BigDecimal(o));
	}

	public void setBigDecimal(PreparedStatement ps, int idx, BigDecimal o)
			throws SQLException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.DECIMAL);
		else
			ps.setBigDecimal(idx, o);
	}

	public void setSerialized(PreparedStatement ps, int idx, Serializable o)
			throws SQLException, IOException {
		if (o == null)
			ps.setNull(idx, java.sql.Types.VARBINARY);
		else {
			Blob blob = new RdbBlob(o);
			ps.setBinaryStream(idx, blob.getBinaryStream(), (int) blob.length());
		}
	}

	public void setArray(PreparedStatement ps, int idx, Object o, int typeCode)
			throws SQLException {
		if (o == null) {
			ps.setNull(idx, Types.ARRAY);
		} else {
			try {
				ps.setArray(idx, new BasicArray(o, typeCode, this, null));
			} catch (RdbAdapterException e) {
				throw new SQLException(e.getMessage());
			}
		}
	}
	
	public void escapeFunctionOpen(StringBuffer sb) {
		sb.append("{fn ");
	}

	public void escapeFunctionClose(StringBuffer sb) {
		sb.append("}");
	}

	public String getConcatExpression(String op1, String op2) {
		StringBuffer sb = new StringBuffer();
		sb.append(CONCAT).append("(").append(op1);
		sb.append(", ").append(op2).append(")");
		return sb.toString();
	}

	public String getFirstLocateExpression(String substring, String instring)
			throws RdbAdapterException {
		StringBuffer sb = new StringBuffer();
		escapeFunctionOpen(sb);
		sb.append("locate(").append(substring).append(", ").append(instring);
		sb.append(")");
		escapeFunctionClose(sb);
		return sb.toString();
	}

	public String getIndexedLocateExpression(String instring,
			String substring,
			String fromIndex) throws RdbAdapterException {
		StringBuffer sb = new StringBuffer();
		sb.append(INDEXEDLOCATE).append("(").append(instring).append(", ");
		sb.append(substring).append(", ").append(fromIndex).append(")");
		return sb.toString();
	}

	public String getLengthOperator() throws RdbAdapterException {
		return "length";
	}

	public String getQuery(String selectClause,
			       List jts,
			       String whereClause,
			       boolean rangeStart,
			       boolean rangeSize) {
		return getQuery(selectClause, getFromClause(jts), whereClause,
				null, null, rangeStart, rangeSize);
	}

	public String getFromClause(List jts) {
		StringBuffer sb = new StringBuffer();
		String sep = "";
		for (int i = 0; i < jts.size(); i++) {
			JoinedTable jt = (JoinedTable) jts.get(i);
			sb.append(sep);
			sep = ", ";
			getFromClause(jt, sb);
		}
		return sb.toString();
	}

	protected void appendClause(StringBuffer sb, String clause,
				    String value) {
		if (value != null && value.length() > 0) {
			sb.append(clause);
			sb.append(value);
		}
	}

	public String getQuery(String selectClause,
			       String fromClause,
			       String whereClause,
			       String orderBy,
			       String groupBy,
			       boolean rangeStart,
			       boolean rangeSize) {
		StringBuffer sb = new StringBuffer();
		sb.append("SELECT ");
        if (rangeStart || rangeSize) {
            modifySelectClauseWithRange(sb, rangeStart, rangeSize);
        }
		sb.append(selectClause);

		appendClause(sb, " FROM ", fromClause);
		appendClause(sb, " WHERE ", whereClause);
		appendClause(sb, " ORDER BY ", orderBy);
		appendClause(sb, " GROUP BY ", groupBy);

        if (rangeStart || rangeSize) {
            modifyQueryWithRange(sb, rangeStart, rangeSize);
        }

		return sb.toString();
	}

	protected void getFromClause(JoinedTable jt, StringBuffer sb) {
		sb.append(jt.tableName);
		if (jt.alias != null && jt.alias.length() > 0) {
			writeTableAlias(jt.alias, sb);
		}
		for (int i = 0; i < jt.joinedTables.size(); i++) {
			getFromClause(jt, (JoinedTable) jt.joinedTables.get(i),
					(JoinedTable.Join) jt.joins.get(i), sb);
		}
	}

	public void writeTableAlias(String alias, StringBuffer sb) {
		sb.append(" as ");
		sb.append(alias);
	}

	public void writeColumnAlias(String alias, StringBuffer sb) {
		sb.append(" as ");
		sb.append(alias);
	}

	public String getColumnAliasExpr(String aliasName) {
		return " as " + aliasName;
	}

	public String handleOrderBy(String query) {
		return query;
	}

	public String getCreateSequence(String seqName)
			throws UnsupportedOperationException {
		return "CREATE SEQUENCE " + seqName;
	}

	public String getDropSequence(String seqName)
		throws UnsupportedOperationException {
		return "DROP SEQUENCE " + seqName;
	}
	
	public String getNextValInSequence(String seqName) {
		return "select nextVal('" + seqName + "')";
	}

	public void getManyNextValInSequence(Connection conn, String seqName, int numberOfNextVal) 
		throws RdbAdapterException {
		StringBuffer sb = new StringBuffer();
		sb.append("select ");
		for (int i = 0; i < numberOfNextVal -1; i++) {
			sb.append("nextVal('" + seqName + "'), ");
		}
		sb.append("nextVal('" + seqName + "')");
		PreparedStatement ps = null;
		try {
			ps = conn.prepareStatement(sb.toString());
			ps.executeQuery();
		} catch (SQLException e) {
			throw new RdbAdapterException(e, "Impossible to allocate " + numberOfNextVal + " identifier"
				+ " on the sequence: " + seqName);
		} finally {
			try {
				if (ps != null) {
					ps.close();
				}
			} catch (SQLException e) {
			}
		}
	}
	
	protected boolean existRelation(Connection connection,
			String relName,
			String[] relationTypes) throws SQLException {
		boolean existtable = false;
		ResultSet rs = null;

		if (relationTypes == null) {
			relationTypes = new String[] { "TABLE", "VIEW" };
		}

		try {
			rs = connection.getMetaData().getTables(null, null, relName,
					relationTypes);
			while (rs.next() && !existtable) {
				existtable = relName.equalsIgnoreCase(rs.getString(3));
			}
		} finally {
			if (rs != null) {
				rs.close();
			}
		}
		return existtable;
	}

	public boolean existTable(Connection connection, String tableName)
			throws SQLException {
		return existRelation(connection, tableName, new String[] { "TABLE" });
	}

	public boolean existView(Connection connection, String viewName)
			throws SQLException {
		return existRelation(connection, viewName, new String[] { "VIEW" });
	}

	public boolean existSequence(Connection connection, String seqName)
			throws SQLException {
		return existTable(connection, seqName);
	}

	protected void getFromClause(JoinedTable jtMain,
				     JoinedTable jt,
				     JoinedTable.Join join,
				     StringBuffer sb) {
                if (join.outer) {
			sb.append(" left outer join ");
                } else {
			sb.append(" join ");
                }
		sb.append(jt.tableName);
		if (jt.alias != null && jt.alias.length() > 0) {
			writeTableAlias(jt.alias, sb);
		}
		sb.append(" on ");
		String sep = "";
		String main = (jtMain.alias != null && jtMain.alias.length() > 0 ? jtMain.alias
				: jtMain.tableName);
		String ext = (jt.alias != null && jt.alias.length() > 0 ? jt.alias
				: jt.tableName);
		for (int i = 0; i < join.colT1.size(); i++) {
			sb.append(sep);
			sep = " AND ";
			sb.append(main);
			sb.append(".");
			sb.append(join.colT1.get(i));
			sb.append(" = ");
			sb.append(ext);
			sb.append(".");
			sb.append(join.colT2.get(i));
		}
		for (int i = 0; i < jt.joinedTables.size(); i++) {
			getFromClause(jt, (JoinedTable) jt.joinedTables.get(i),
					(JoinedTable.Join) jt.joins.get(i), sb);
		}
	}

	public static char[] getCharArray(Object o) throws IOException,
			RdbAdapterException {
		if (o instanceof java.io.Reader) {
			java.io.Reader reader = (java.io.Reader) o;
			int size = 0;
			long skipped = 0;
			while ((skipped = reader.skip(50)) == 50) {
				size += 50;
			}
			size += (int) skipped;
			char[] res = new char[size];
			reader.reset();
			reader.read(res);
			return res;
		} else if (o instanceof String) {
			return ((String) o).toCharArray();
		} else {
			throw new RdbAdapterException("Unknown type to convert to char[]");
		}
	}

	public String getSubqueryAlias() {
		return "foo";
	}

	/**
	 * CREATE [ TEMPORARY | TEMP ] SEQUENCE name [ INCREMENT [ BY ] increment ] [
	 * MINVALUE minvalue | NO MINVALUE ] [ MAXVALUE maxvalue | NO MAXVALUE ] [
	 * START [ WITH ] start ] [ CACHE cache ] [ [ NO ] CYCLE ]
	 * 
	 * @param seqName
	 * @param startid
	 * @param inc
	 */
	public String getCreateSequence(String seqName,
			Integer startid,
			Integer inc,
			Integer cache) throws UnsupportedOperationException {
		StringBuffer sb = new StringBuffer("CREATE SEQUENCE ");
		sb.append(seqName);
		if (inc != null) {
			sb.append(" ");
			sb.append(getINCREMENTkeyWord());
			sb.append(" ");
			sb.append(inc);
		}
		if (startid != null) {
			sb.append(" ");
			sb.append(getSTARTkeyWord());
			sb.append(" ");
			sb.append(startid);
		}
		if (cache != null) {
			sb.append(" ");
			sb.append(getCACHEkeyWord());
			sb.append(" ");
			sb.append(cache);
		}
		return sb.toString();
	}

	public void setLogger(Logger _logger) {
	    this.logger = _logger;
	}
	
	protected String getINCREMENTkeyWord() {
		return "INCREMENT BY";
	}

	protected String getSTARTkeyWord() {
		return "START WITH";
	}

	protected String getCACHEkeyWord() {
		return "CACHE";
	}

	protected void modifySelectClauseWithRange(StringBuffer sb,
	        boolean rangeStart,
	        boolean rangeSize) {
	    if (logger != null) {
	        logger.log(BasicLevel.WARN, "Range queries not suppoted by this adapter (" + this.name + ") , leaving select clause unchanged.");
	    }
	}

	protected void modifyQueryWithRange(StringBuffer sb,
	        boolean rangeStart,
	        boolean rangeSize) {
	    if (logger != null) {
	        logger.log(BasicLevel.WARN, "Range queries not suppoted by this adapter (" + this.name + ") , leaving query unchanged.");
	    }
	}

    public int getRangeParametersAtStart() {
	    if (logger != null) {
	        logger.log(BasicLevel.WARN, "Range queries not suppoted by this adapter (" + this.name + ") , leaving query unchanged.");
	    }
        return RANGEUNSUPPORTED;
    }
    
    
	public boolean supportArray(int elemTypeCode) {
		return true;
	}
	
	//------------ INDEX ADAPTER INTERFACE -------------------//
	
	public String getDropIndex(String name, String tableName)
	throws UnsupportedOperationException {
		return "DROP INDEX " + name;
	}
	
	public String getCreateIndex(String name, String tableName, List fieldNames, boolean unique)
	throws UnsupportedOperationException {
		StringBuffer sb = new StringBuffer("CREATE");
		if (unique) {
			sb.append(" UNIQUE");
		}
		sb.append(" INDEX " + name + " ON " + tableName + "(");
		for (int i = 0; i < fieldNames.size(); i++) {
			if (i == 0) {
				sb.append(fieldNames.get(i));
			} else {
				sb.append(","+fieldNames.get(i));
			}
		}
		sb.append(")");
		return sb.toString(); 
	}
}