/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2005 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.adapter;

import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import java.sql.Types;
import java.util.Date;

/**
 * MySql 4.0.12
 * 
 * @author S.Chassande-Barrioz
 */
public class MysqlAdapter extends BasicRdbAdapter {

    private final static String MYSQLINDEXEDLOCATE = "locate";

    public MysqlAdapter() {
        super("mysql");
    }

    public boolean supportBatchPreparedStatement() {
        return false;
    }

    /**
     * This calculates the size of a result set. If this feature is not
     * supported by a database or its jdbc driver, this method returns -1.
     * 
     * @param rs
     *            the result set which the size must be returned
     */
    public int fetchResultSetSize(ResultSet rs) throws SQLException {
        return -1;
    }

    // MAX LENGTH OF VARCHAR TYPE MYSQL IS LIMITED to 255
    private int varcharSize = 254;

    private int varcharSizeInPk = 200;

    public int varbinarySize = 254;

    public int varbinarySizeInPk = 200;

    public void setVarcharSize(int s) {
        varcharSize = s;
    }

    public void setVarcharSizeInPk(int s) {
        varcharSizeInPk = s;
    }

    public String getSqlType(int typeCode, boolean usedInPK, int size, int scale)
            throws RdbAdapterException {
        // ALL COLUMN IN PK SHOULD BE DECLARED NOT NULL
        int realsize = size;
        if (BasicArray.isArray(typeCode)) {
            throw new UnsupportedOperationException("Mysql does not support java.sql.Types.ARRAY");
        }
        switch (typeCode) {
        case TYPECODE_STRING:
            if (realsize == NOSIZE) {
                realsize = (usedInPK ? varcharSizeInPk : varcharSize);
            }
            return (usedInPK ? "VARCHAR(" + realsize + ")  NOT NULL"
                    : "LONGTEXT ");
        case TYPECODE_SERIALIZED:
            return "LONGTEXT";
        case TYPECODE_BIGINTEGER:
            if (size == NOSIZE) {
                size = 127;
            }
            return "DECIMAL(" + size + ",0)";
        case TYPECODE_BIGDECIMAL:
            if (size != NOSIZE) {
                if (scale != NOSIZE) {
                    return "DECIMAL(" + size + ", " + scale + ")";
                } else {
                    return "DECIMAL(" + size + ")";
                }
            } else {
                return "DECIMAL(127,127)";
            }
        case TYPECODE_BYTEARRAY:
            if (realsize == NOSIZE) {
                realsize = (usedInPK ? varbinarySizeInPk : varbinarySize);
            }
            return (usedInPK ? "VARCHAR(" + realsize + ") BINARY NOT NULL"
                    : "LONGTEXT ");
        case TYPECODE_CHARARRAY:
            if (realsize == NOSIZE) {
                realsize = (usedInPK ? varbinarySizeInPk : varbinarySize);
            }
            return (usedInPK ? "VARCHAR(" + realsize + ") NOT NULL"
                    : "LONGTEXT ");
        case TYPECODE_DATE:
            return "DATETIME";
        default:
            String res = super.getSqlType(typeCode, usedInPK, size, scale);
            if (usedInPK) {
                res += " NOT NULL";
            }
            return res;
        }
    }

    public int getSqlTypeCode(int typeCode, String sqlColumnType)
            throws RdbAdapterException {
        if (BasicArray.isArray(typeCode)) {
            throw new UnsupportedOperationException("Mysql does not support java.sql.Types.ARRAY");
        }
        switch (typeCode) {
        case TYPECODE_DATE:
            if (sqlColumnType == null) {
                return Types.TIMESTAMP;
            } else {
                String str = sqlColumnType.toUpperCase();
                if (str.startsWith("DATE")) {
                    return Types.DATE;
                } else if (str.startsWith("TIMESTAMP")) {
                    return Types.TIMESTAMP;
                } else if (str.startsWith("TIME")) {
                    return Types.TIME;
                } else {
                    return Types.TIMESTAMP;
                }
            }
        default:
            return super.getSqlTypeCode(typeCode, sqlColumnType);
        }
    }

    public char getChar(ResultSet rs, int idx, char nullValue)
            throws SQLException {
        String res = rs.getString(idx);
        if (rs.wasNull() || res == null) {
            return nullValue;
        }
        if (res.length() == 0) {
            return ' ';
        } else {
            return res.charAt(0);
        }
    }

    public Character getOchar(ResultSet rs, int idx, Character nullValue)
            throws SQLException {
        String res = rs.getString(idx);
        if (rs.wasNull() || res == null) {
            return nullValue;
        }
        if (res.length() == 0) {
            return new Character(' ');
        } else {
            return new Character(res.charAt(0));
        }
    }

    public Date getDate(ResultSet rs, int idx, String sqlColumnType,
            Date nullValue) throws SQLException {
        Date res;
        if (sqlColumnType == null) {
            res = rs.getTimestamp(idx);
        } else {
            String str = sqlColumnType.toUpperCase();
            if (str.startsWith("TIMESTAMP") || str.equals("DATETIME")) {
                res = rs.getTimestamp(idx);
            } else if (str.startsWith("DATE")) {
                res = rs.getDate(idx);
            } else if (str.startsWith("TIME")) {
                res = rs.getTime(idx);
            } else {
                res = rs.getTimestamp(idx);
            }
        }
        return (rs.wasNull() ? nullValue : res);
    }

    public String getIndexedLocateExpression(String instring, String substring,
            String fromIndex) {
        return MYSQLINDEXEDLOCATE + "(" + substring + ", " + instring + ", "
                + fromIndex + ")";
    }

    public void setDate(PreparedStatement ps, int idx, Date o, String columnType)
            throws SQLException {
        if (columnType == null) {
            //default is a Timestamp
            if (o == null)
                ps.setTimestamp(idx, new java.sql.Timestamp(0));
            else if (o instanceof java.sql.Timestamp) {
                ps.setTimestamp(idx, (java.sql.Timestamp) o);
            } else {
                ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
            }
        } else if (columnType.startsWith("TIMESTAMP")
                || columnType.equals("DATETIME")) {
            if (o == null)
                ps.setTimestamp(idx, new java.sql.Timestamp(0));
            else if (o instanceof java.sql.Timestamp) {
                ps.setTimestamp(idx, (java.sql.Timestamp) o);
            } else {
                ps.setTimestamp(idx, new java.sql.Timestamp(o.getTime()));
            }
        } else if (columnType.startsWith("TIME")) {
            if (o == null)
                ps.setTime(idx, new java.sql.Time(0));
            else if (o instanceof java.sql.Time) {
                ps.setTime(idx, (java.sql.Time) o);
            } else {
                ps.setTime(idx, new java.sql.Time(o.getTime()));
            }
        } else if (columnType.startsWith("DATE")) {
            if (o == null)
                ps.setDate(idx, new java.sql.Date(0));
            else if (o instanceof java.sql.Date) {
                ps.setDate(idx, (java.sql.Date) o);
            } else {
                ps.setDate(idx, new java.sql.Date(o.getTime()));
            }
        }
    }

    public String getCreateSequence(String seqName)
            throws UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "Mysql does not support SEQUENCE");
    }

    public String getCreateSequence(String seqName, Integer startid,
            Integer inc, Integer cache) {
        throw new UnsupportedOperationException(
                "Mysql does not support SEQUENCE");
    }

    public String getDropSequence(String seqName)
    throws UnsupportedOperationException {
    	throw new UnsupportedOperationException(
        	"Mysql does not support SEQUENCE");
    }
    
    protected void modifySelectClauseWithRange(StringBuffer sb, boolean rangeStart,
            boolean rangeSize) {
        //do nothing to avoid the warning message
    }

    protected void modifyQueryWithRange(StringBuffer sb, boolean rangeStart,
            boolean rangeSize) {
        if (rangeSize || rangeStart) {
	        if (rangeStart) {
	            sb.append (" LIMIT ?, ");
	        } else {
	            sb.append(" LIMIT 0, ");
	        }
	        if (rangeSize) {
	            sb.append(" ? ");
	        } else {
	            // this idiom for skipping the first n rows is
	            // copied straight out of the mysql
	            // documentation
	            sb.append(" 18446744073709551615 ");
	        }
        }
    }

    public int getRangeParametersAtStart() {
        return RANGEATEND;
    }
    
    public String getSubqueryAlias() {
        return "foo";
    }
    
	public Object getArray(ResultSet rs, int idx, Object nullValue, int typeCode)
			throws SQLException {
        throw new UnsupportedOperationException("Mysql does not support java.sql.Types.ARRAY");
	}
	
	public String getArrayValueAsSQLString(Object array, int typeCode,
			String begin, String realsep, String end) {
        throw new UnsupportedOperationException("Mysql does not support java.sql.Types.ARRAY");
	}
	
	public boolean supportArray(int elemTypeCode) {
		return false;
	}	
	
	//-------------- INTERFACE INDEX ADAPTER ------------------//
	public String getDropIndex(String name, String tableName)
	throws UnsupportedOperationException {
		return "DROP INDEX " + name + " ON " + tableName;
	}
}